@@ -19,10 +19,10 @@ from __future__ import unicode_literals
 
 __intname__ = "command_runner"
 __author__ = "Orsiris de Jong"
-__copyright__ = "Copyright (C) 2015-2022 Orsiris de Jong"
+__copyright__ = "Copyright (C) 2015-2025 Orsiris de Jong for NetInvent"
 __licence__ = "BSD 3 Clause"
-__version__ = "1.4.0"
-__build__ = "2022053001"
+__version__ = "1.7.4"
+__build__ = "2025052301"
 __compat__ = "python2.7+"
 
 import io
@@ -34,11 +34,91 @@ from datetime import datetime
 from logging import getLogger
 from time import sleep
 
+
+# Avoid checking os type numerous times
+os_name = os.name
+
+
+# Don't bother with an ImportError since we need command_runner to work without dependencies
 try:
     import psutil
-except ImportError:
-    # Don't bother with an error since we need command_runner to work without dependencies
-    pass
+
+    # Also make sure we directly import priority classes so we can reuse them
+    if os_name == "nt":
+        from psutil import (
+            # ABOVE_NORMAL_PRIORITY_CLASS,
+            BELOW_NORMAL_PRIORITY_CLASS,
+            HIGH_PRIORITY_CLASS,
+            IDLE_PRIORITY_CLASS,
+            NORMAL_PRIORITY_CLASS,
+            REALTIME_PRIORITY_CLASS,
+        )
+        from psutil import (
+            IOPRIO_HIGH,
+            IOPRIO_NORMAL,
+            IOPRIO_LOW,
+            # IOPRIO_VERYLOW,
+        )
+    else:
+        from psutil import (
+            IOPRIO_CLASS_BE,
+            IOPRIO_CLASS_IDLE,
+            # IOPRIO_CLASS_NONE,
+            IOPRIO_CLASS_RT,
+        )
+except (ImportError, AttributeError):
+    if os_name == "nt":
+        BELOW_NORMAL_PRIORITY_CLASS = 16384
+        HIGH_PRIORITY_CLASS = 128
+        NORMAL_PRIORITY_CLASS = 32
+        REALTIME_PRIORITY_CLASS = 256
+        IDLE_PRIORITY_CLASS = 64
+        IOPRIO_HIGH = 3
+        IOPRIO_NORMAL = 2
+        IOPRIO_LOW = 1
+    else:
+        IOPRIO_CLASS_IDLE = 3
+        IOPRIO_CLASS_BE = 2
+        IOPRIO_CLASS_RT = 1
+
+
+# Python 2.7 does not have priorities defined in subprocess module, but psutil has
+# Since Windows and Linux use different possible values, let's simplify things by
+# allowing 5 process priorities: verylow (idle), low, normal, high, rt
+# and 3 process io priorities: low, normal, high
+# For IO, rt == high
+if os_name == "nt":
+    PRIORITIES = {
+        "process": {
+            "verylow": IDLE_PRIORITY_CLASS,
+            "low": BELOW_NORMAL_PRIORITY_CLASS,
+            "normal": NORMAL_PRIORITY_CLASS,
+            "high": HIGH_PRIORITY_CLASS,
+            "rt": REALTIME_PRIORITY_CLASS,
+        },
+        "io": {
+            "low": IOPRIO_LOW,
+            "normal": IOPRIO_NORMAL,
+            "high": IOPRIO_HIGH,
+        },
+    }
+else:
+    PRIORITIES = {
+        "process": {
+            "verylow": 20,
+            "low": 15,
+            "normal": 0,
+            "high": -15,
+            "rt": -20,
+        },
+        "io": {
+            "low": IOPRIO_CLASS_IDLE,
+            "normal": IOPRIO_CLASS_BE,
+            "high": IOPRIO_CLASS_RT,
+        },
+    }
+
+
 try:
     import signal
 except ImportError:
@@ -53,7 +133,7 @@ import threading
 
 # Python 2.7 compat fixes (missing typing)
 try:
-    from typing import Union, Optional, List, Tuple, NoReturn, Any, Callable
+    from typing import Union, Optional, List, Tuple, Any, Callable
 except ImportError:
     pass
 
@@ -65,9 +145,15 @@ except ImportError:
     # Python 2.7 just won't have concurrent.futures, so we just declare threaded and wraps in order to
     # avoid NameError
     def threaded(fn):
+        """
+        Simple placeholder for python 2.7
+        """
         return fn
 
     def wraps(fn):
+        """
+        Simple placeholder for python 2.7
+        """
         return fn
 
 
@@ -89,11 +175,16 @@ except AttributeError:
         Basic redeclaration when subprocess.TimeoutExpired does not exist, python <= 3.3
         """
 
-        def __init__(self, cmd, timeout, output=None, stderr=None):
+        def __init__(self, cmd, timeout, output=None, stderr=None, *args, **kwargs):
             self.cmd = cmd
             self.timeout = timeout
             self.output = output
             self.stderr = stderr
+            try:
+                super().__init__(*args, **kwargs)
+            except TypeError:
+                # python 2.7 needs super(Baseclass, self)
+                super(BaseException, self).__init__(*args, **kwargs)
 
         def __str__(self):
             return "Command '%s' timed out after %s seconds" % (self.cmd, self.timeout)
@@ -114,8 +205,13 @@ class InterruptGetOutput(BaseException):
     Make sure we get the current output when process is stopped mid-execution
     """
 
-    def __init__(self, output):
+    def __init__(self, output, *args, **kwargs):
         self._output = output
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(BaseException, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -127,8 +223,13 @@ class KbdInterruptGetOutput(InterruptGetOutput):
     Make sure we get the current output when KeyboardInterrupt is made
     """
 
-    def __init__(self, output):
+    def __init__(self, output, *args, **kwargs):
         self._output = output
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(InterruptGetOutput, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -140,8 +241,13 @@ class StopOnInterrupt(InterruptGetOutput):
     Make sure we get the current output when optional stop_on function execution returns True
     """
 
-    def __init__(self, output):
+    def __init__(self, output, *args, **kwargs):
         self._output = output
+        try:
+            super().__init__(*args, **kwargs)
+        except TypeError:
+            # python 2.7 needs super(Baseclass, self)
+            super(InterruptGetOutput, self).__init__(*args, **kwargs)
 
     @property
     def output(self):
@@ -154,6 +260,18 @@ def call_with_future(fn, future, args, kwargs):
     Threading a function with return info using Future
     from https://stackoverflow.com/a/19846691/2635443
 
+    Example:
+
+    @threaded
+    def somefunc(arg):
+        return 'arg was %s' % arg
+
+
+    thread = somefunc('foo')
+    while thread.done() is False:
+        time.sleep(1)
+
+    print(thread.result())
     """
     try:
         result = fn(*args, **kwargs)
@@ -174,10 +292,14 @@ def threaded(fn):
 
     @wraps(fn)
     def wrapper(*args, **kwargs):
+        if kwargs.pop("__no_threads", False):
+            return fn(*args, **kwargs)
         future = Future()
-        threading.Thread(
+        thread = threading.Thread(
             target=call_with_future, args=(fn, future, args, kwargs)
-        ).start()
+        )
+        thread.daemon = True
+        thread.start()
         return future
 
     return wrapper
@@ -190,6 +312,80 @@ logger = getLogger(__intname__)
 PIPE = subprocess.PIPE
 
 
+def _validate_process_priority(
+    priority,  # type: Union[int, str]
+):
+    # type: (...) -> int
+    """
+    Check if priority int is valid
+    """
+
+    def _raise_prio_error(priority, reason):
+        raise ValueError(
+            "Priority not valid ({}): {}. Please use one of {}".format(
+                reason, priority, ", ".join(list(PRIORITIES["process"].keys()))
+            )
+        )
+
+    if isinstance(priority, int):
+        if os_name == "nt":
+            _raise_prio_error(priority, "windows does not accept ints as priority")
+        if -20 <= priority <= 20:
+            _raise_prio_error(priority, "priority out of range")
+    elif isinstance(priority, str):
+        try:
+            priority = PRIORITIES["process"][priority.lower()]
+        except KeyError:
+            _raise_prio_error(priority, "priority does not exist")
+    return priority
+
+
+def _set_priority(
+    pid,  # type: int
+    priority,  # type: Union[int, str]
+    priority_type,  # type: str
+):
+    """
+    Set process and / or io prioritie
+    """
+    priority = priority.lower()
+
+    if priority_type == "process":
+        _priority = _validate_process_priority(priority)
+        psutil.Process(pid).nice(_priority)
+    elif priority_type == "io":
+        valid_io_priorities = list(PRIORITIES["io"].keys())
+        if priority not in valid_io_priorities:
+            raise ValueError(
+                "Bogus {} priority given: {}. Please use one of {}".format(
+                    priority_type, priority, ", ".join(valid_io_priorities)
+                )
+            )
+        psutil.Process(pid).ionice(PRIORITIES[priority_type][priority])
+    else:
+        raise ValueError("Bogus priority type given.")
+
+
+def set_priority(
+    pid,  # type: int
+    priority,  # type: Union[int, str]
+):
+    """
+    Shorthand for _set_priority
+    """
+    _set_priority(pid, priority, "process")
+
+
+def set_io_priority(
+    pid,  # type: int
+    priority,  # type: str
+):
+    """
+    Shorthand for _set_priority
+    """
+    _set_priority(pid, priority, "io")
+
+
 def to_encoding(
     process_output,  # type: Union[str, bytes]
     encoding,  # type: Optional[str]
@@ -198,7 +394,7 @@ def to_encoding(
     # type: (...) -> str
     """
     Convert bytes output to string and handles conversion errors
-    Varation of ofunctions.string_handling.safe_string_convert
+    Variation of ofunctions.string_handling.safe_string_convert
     """
 
     if not encoding:
@@ -214,7 +410,10 @@ def to_encoding(
                 process_output = process_output.decode(encoding, errors="ignore")
             except (ValueError, TypeError):
                 # What happens when str cannot be concatenated
-                logger.debug("Output cannot be captured {}".format(process_output))
+                logger.error("Output cannot be captured {}".format(process_output))
+    elif process_output is None:
+        # We deal with strings. Alter output string to avoid NoneType errors
+        process_output = ""
     return process_output
 
 
@@ -227,7 +426,7 @@ def kill_childs_mod(
     """
     Inline version of ofunctions.kill_childs that has no hard dependency on psutil
 
-    Kills all childs of pid (current pid can be obtained with os.getpid())
+    Kills all children of pid (current pid can be obtained with os.getpid())
     If no pid given current pid is taken
     Good idea when using multiprocessing, is to call with atexit.register(ofunctions.kill_childs, os.getpid(),)
 
@@ -311,7 +510,7 @@ def kill_childs_mod(
                     pid, 15
                 )  # 15 being signal.SIGTERM or SIGKILL depending on the platform
             except OSError as exc:
-                if os.name == "nt":
+                if os_name == "nt":
                     # We'll do an ugly hack since os.kill() has some pretty big caveats on Windows
                     # especially for Python 2.7 where we can get Access Denied
                     os.system("taskkill /F /pid {}".format(pid))
@@ -336,19 +535,26 @@ def kill_childs_mod(
 
 def command_runner(
     command,  # type: Union[str, List[str]]
-    valid_exit_codes=None,  # type: Optional[List[int]]
+    valid_exit_codes=False,  # type: Union[List[int], bool]
     timeout=3600,  # type: Optional[int]
     shell=False,  # type: bool
     encoding=None,  # type: Optional[Union[str, bool]]
+    stdin=None,  # type: Optional[Union[int, str, Callable, queue.Queue]]
     stdout=None,  # type: Optional[Union[int, str, Callable, queue.Queue]]
     stderr=None,  # type: Optional[Union[int, str, Callable, queue.Queue]]
+    no_close_queues=False,  # type: Optional[bool]
     windows_no_window=False,  # type: bool
     live_output=False,  # type: bool
     method="monitor",  # type: str
     check_interval=0.05,  # type: float
     stop_on=None,  # type: Callable
+    on_exit=None,  # type: Callable
     process_callback=None,  # type: Callable
     split_streams=False,  # type: bool
+    silent=False,  # type: bool
+    priority=None,  # type: Union[int, str]
+    io_priority=None,  # type: str
+    heartbeat=0,  # type: int
     **kwargs  # type: Any
 ):
     # type: (...) -> Union[Tuple[int, Optional[Union[bytes, str]]], Tuple[int, Optional[Union[bytes, str]], Optional[Union[bytes, str]]]]
@@ -359,8 +565,8 @@ def command_runner(
 
     Also allows a list of various valid exit codes (ie no error when exit code = arbitrary int)
 
-    command should be a list of strings, eg ['ping', '127.0.0.1', '-c 2']
-    command can also be a single string, ex 'ping 127.0.0.1 -c 2' if shell=True or if os is Windows
+    command should be a list of strings, eg ['ping', '-c 2', '127.0.0.1']
+    command can also be a single string, ex 'ping -c 2 127.0.0.1' if shell=True or if os is Windows
 
     Accepts all of subprocess.popen arguments
 
@@ -377,19 +583,24 @@ def command_runner(
 
     stop_on is an optional function that will stop execution if function returns True
 
+    priority and io_priority can be set to 'low', 'normal' or 'high'
+    priority may also be an int from -20 to 20 on Unix
+
+    heartbeat will log a line every heartbeat seconds informing that we're still alive
+
     Returns a tuple (exit_code, output)
     """
 
     # Choose default encoding when none set
     # cp437 encoding assures we catch most special characters from cmd.exe
     # Unless encoding=False in which case nothing gets encoded except Exceptions and logger strings for Python 2
-    error_encoding = "cp437" if os.name == "nt" else "utf-8"
-    if encoding is not False:
+    error_encoding = "cp437" if os_name == "nt" else "utf-8"
+    if encoding is None:
         encoding = error_encoding
 
     # Fix when unix command was given as single string
     # This is more secure than setting shell=True
-    if os.name == "posix":
+    if os_name == "posix":
         if not shell and isinstance(command, str):
             command = shlex.split(command)
         elif shell and isinstance(command, list):
@@ -406,7 +617,7 @@ def command_runner(
         windows_no_window
         and sys.version_info[0] >= 3
         and sys.version_info[1] >= 7
-        and os.name == "nt"
+        and os_name == "nt"
     ):
         # Disable the following pylint error since the code also runs on nt platform, but
         # triggers an error on Unix
@@ -488,7 +699,7 @@ def command_runner(
 
         # We also need to check that stream has readline, in case we're writing to files instead of PIPE
 
-        # Another magnificient python 2.7 fix
+        # Another magnificent python 2.7 fix
         # So we need to convert sentinel_char which would be unicode because of unicode_litterals
         # to str which is the output format from stream.readline()
 
@@ -512,6 +723,37 @@ def command_runner(
                 return output_stderr
             return None
 
+    def _heartbeat_thread(
+        process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
+        heartbeat,  # type: int
+    ):
+        begin_time = datetime.now()
+        while True:
+            elapsed_time = int((datetime.now() - begin_time).total_seconds())
+            if elapsed_time > heartbeat and elapsed_time % heartbeat == 0:
+                logger.info("Still running command after %s seconds" % elapsed_time)
+            if process.poll() is not None:
+                break
+            sleep(1)
+
+    def heartbeat_thread(
+        process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
+        heartbeat,  # type: int
+    ):
+        """
+        Just a shorthand to run the heartbeat thread
+        """
+        if heartbeat:
+            heartbeat_thread = threading.Thread(
+                target=_heartbeat_thread,
+                args=(
+                    process,
+                    heartbeat,
+                ),
+            )
+            heartbeat_thread.daemon = True
+            heartbeat_thread.start()
+
     def _poll_process(
         process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
         timeout,  # type: int
@@ -537,9 +779,8 @@ def command_runner(
             # type: (...) -> None
             """
             Simple subfunction to check whether timeout is reached
-            Since we check this alot, we put it into a function
+            Since we check this a lot, we put it into a function
             """
-
             if timeout and (datetime.now() - begin_time).total_seconds() > timeout:
                 kill_childs_mod(process.pid, itself=True, soft_kill=False)
                 raise TimeoutExpired(
@@ -550,6 +791,9 @@ def command_runner(
                 raise StopOnInterrupt(_get_error_output(output_stdout, output_stderr))
 
         begin_time = datetime.now()
+
+        heartbeat_thread(process, heartbeat)
+
         if encoding is False:
             output_stdout = output_stderr = b""
         else:
@@ -631,8 +875,7 @@ def command_runner(
             exit_code = process.poll()
             if split_streams:
                 return exit_code, output_stdout, output_stderr
-            else:
-                return exit_code, output_stdout
+            return exit_code, output_stdout
 
         except KeyboardInterrupt:
             raise KbdInterruptGetOutput(_get_error_output(output_stdout, output_stderr))
@@ -661,7 +904,7 @@ def command_runner(
                 break
             if process.poll() is not None:
                 break
-            # We definitly need some sleep time here or else we will overload CPU
+            # We definitely need some sleep time here or else we will overload CPU
             sleep(check_interval)
 
     def _monitor_process(
@@ -689,6 +932,8 @@ def command_runner(
         thread.daemon = True  # was setDaemon(True) which has been deprecated
         thread.start()
 
+        heartbeat_thread(process, heartbeat)
+
         if encoding is False:
             output_stdout = output_stderr = b""
             output_stdout_end = output_stderr_end = b""
@@ -743,12 +988,11 @@ def command_runner(
                 raise TimeoutExpired(
                     process, timeout, _get_error_output(output_stdout, output_stderr)
                 )
-            elif must_stop["value"] == "S":
+            if must_stop["value"] == "S":
                 raise StopOnInterrupt(_get_error_output(output_stdout, output_stderr))
             if split_streams:
                 return exit_code, output_stdout, output_stderr
-            else:
-                return exit_code, output_stdout
+            return exit_code, output_stdout
         except KeyboardInterrupt:
             raise KbdInterruptGetOutput(_get_error_output(output_stdout, output_stderr))
 
@@ -774,14 +1018,29 @@ def command_runner(
 
         # Python >= 3.3 has SubProcessError(TimeoutExpired) class
         # Python >= 3.6 has encoding & error arguments
+        # Python >= 3.7 has creationflags arguments for process priority under windows
         # universal_newlines=True makes netstat command fail under windows
         # timeout does not work under Python 2.7 with subprocess32 < 3.5
         # decoder may be cp437 or unicode_escape for dos commands or utf-8 for powershell
+
+        if priority:
+            process_prio = _validate_process_priority(priority)
+            # Don't bother to make pylint go crazy on Windows missing os.nice()
+            # pylint: disable=E1101
+            if os_name == "nt" and sys.version_info >= (3, 7):
+                creationflags |= process_prio
+
+            # Actually we don't want preexec_fn since it's not thread safe, neither supported
+            # in subinterpreters. We'll use set_priority() once the process is spawned
+            # else:
+            #     kwargs["preexec_fn"] = lambda: os.nice(process_prio)
+
         # Disabling pylint error for the same reason as above
         # pylint: disable=E1123
         if sys.version_info >= (3, 6):
             process = subprocess.Popen(
                 command,
+                stdin=stdin,
                 stdout=_stdout,
                 stderr=_stderr,
                 shell=shell,
@@ -796,6 +1055,7 @@ def command_runner(
         else:
             process = subprocess.Popen(
                 command,
+                stdin=stdin,
                 stdout=_stdout,
                 stderr=_stderr,
                 shell=shell,
@@ -806,6 +1066,51 @@ def command_runner(
                 **kwargs
             )
 
+        # Set process priority if not set earlier by creationflags
+        if priority and (
+            os_name != "nt" or (sys.version_info < (3, 7) and os_name == "nt")
+        ):
+            try:
+                try:
+                    set_priority(process.pid, priority)
+                except psutil.AccessDenied as exc:
+                    logger.warning(
+                        "Cannot set process priority {}: {}. Access denied.".format(
+                            priority, exc
+                        )
+                    )
+                    logger.debug("Trace:", exc_info=True)
+                except Exception as exc:
+                    logger.warning(
+                        "Cannot set process priority {}: {}".format(priority, exc)
+                    )
+                    logger.debug("Trace:", exc_info=True)
+            except NameError:
+                logger.warning(
+                    "Cannot set process priority. No psutil module installed."
+                )
+                logger.debug("Trace:", exc_info=True)
+        # Set io priority if given
+        if io_priority:
+            try:
+                try:
+                    set_io_priority(process.pid, io_priority)
+                except psutil.AccessDenied as exc:
+                    logger.warning(
+                        "Cannot set io priority {}: {} Access denied.".format(
+                            io_priority, exc
+                        )
+                    )
+                    logger.debug("Trace:", exc_info=True)
+                except Exception as exc:
+                    logger.warning(
+                        "Cannot set io priority {}: {}".format(io_priority, exc)
+                    )
+                    logger.debug("Trace:", exc_info=True)
+                    raise
+            except NameError:
+                logger.warning("Cannot set io priority. No psutil module installed.")
+
         try:
             # let's return process information if callback was given
             if callable(process_callback):
@@ -845,9 +1150,7 @@ def command_runner(
                 _stdout.write(output_stderr.encode(encoding, errors=errors))
 
         logger.debug(
-            'Command "{}" returned with exit code "{}". Command output was:'.format(
-                command, exit_code
-            )
+            'Command "{}" returned with exit code "{}"'.format(command, exit_code)
         )
     except subprocess.CalledProcessError as exc:
         exit_code = exc.returncode
@@ -855,23 +1158,28 @@ def command_runner(
             output_stdout = exc.output
         except AttributeError:
             output_stdout = "command_runner: Could not obtain output from command."
-        if exit_code in valid_exit_codes if valid_exit_codes is not None else [0]:
-            logger.debug(
-                'Command "{}" returned with exit code "{}". Command output was:'.format(
-                    command, exit_code
+
+        logger_fn = logger.error
+        valid_exit_codes_msg = ""
+        if valid_exit_codes:
+            if valid_exit_codes is True or exit_code in valid_exit_codes:
+                logger_fn = logger.info
+                valid_exit_codes_msg = " allowed"
+
+        if not silent:
+            logger_fn(
+                'Command "{}" failed with{} exit code "{}"'.format(
+                    command,
+                    valid_exit_codes_msg,
+                    exc.returncode,
                 )
             )
-        logger.error(
-            'Command "{}" failed with exit code "{}". Command output was:'.format(
-                command, exc.returncode
-            )
-        )
-        logger.error(output_stdout)
     except FileNotFoundError as exc:
         message = 'Command "{}" failed, file not found: {}'.format(
             command, to_encoding(exc.__str__(), error_encoding, errors)
         )
-        logger.error(message)
+        if not silent:
+            logger.error(message)
         if stdout_destination == "file":
             _stdout.write(message.encode(error_encoding, errors=errors))
         exit_code, output_stdout = (-253, message)
@@ -881,41 +1189,46 @@ def command_runner(
         message = 'Command "{}" failed because of OS: {}'.format(
             command, to_encoding(exc.__str__(), error_encoding, errors)
         )
-        logger.error(message)
+        if not silent:
+            logger.error(message)
         if stdout_destination == "file":
             _stdout.write(message.encode(error_encoding, errors=errors))
         exit_code, output_stdout = (-253, message)
     except TimeoutExpired as exc:
         message = 'Timeout {} seconds expired for command "{}" execution. Original output was: {}'.format(
-            timeout, command, exc.output
+            timeout, command, to_encoding(exc.output, error_encoding, errors)[-1000:]
         )
-        logger.error(message)
+        if not silent:
+            logger.error(message)
         if stdout_destination == "file":
             _stdout.write(message.encode(error_encoding, errors=errors))
         exit_code, output_stdout = (-254, message)
     except StopOnInterrupt as exc:
         message = "Command {} was stopped because stop_on function returned True. Original output was: {}".format(
-            command, to_encoding(exc.output, error_encoding, errors)
+            command, to_encoding(exc.output, error_encoding, errors)[-1000:]
         )
-        logger.info(message)
+        if not silent:
+            logger.info(message)
         if stdout_destination == "file":
             _stdout.write(message.encode(error_encoding, errors=errors))
         exit_code, output_stdout = (-251, message)
     except ValueError as exc:
         message = to_encoding(exc.__str__(), error_encoding, errors)
-        logger.error(message, exc_info=True)
+        if not silent:
+            logger.error(message, exc_info=True)
         if stdout_destination == "file":
             _stdout.write(message)
         exit_code, output_stdout = (-250, message)
     # We need to be able to catch a broad exception
     # pylint: disable=W0703
     except Exception as exc:
-        logger.error(
-            'Command "{}" failed for unknown reasons: {}'.format(
-                command, to_encoding(exc.__str__(), error_encoding, errors)
-            ),
-            exc_info=True,
-        )
+        if not silent:
+            logger.error(
+                'Command "{}" failed for unknown reasons: {}'.format(
+                    command, to_encoding(exc.__str__(), error_encoding, errors)
+                ),
+                exc_info=True,
+            )
         exit_code, output_stdout = (
             -255,
             to_encoding(exc.__str__(), error_encoding, errors),
@@ -926,24 +1239,13 @@ def command_runner(
         if stderr_destination == "file":
             _stderr.close()
 
-    logger.debug(
-        "STDOUT: " + to_encoding(output_stdout, error_encoding, errors)
-        if output_stdout
-        else "None"
-    )
-    if stderr_destination not in ["stdout", None]:
-        logger.debug(
-            "STDERR: " + to_encoding(output_stderr, error_encoding, errors)
-            if output_stderr
-            else "None"
-        )
-
-    # Make sure we send a simple queue end before leaving to make any queue read process will stop regardless
+    # Make sure we send a simple queue end before leaving to make sure any queue read process will stop regardless
     # of command_runner state (useful when launching with queue and method poller which isn't supposed to write queues)
-    if stdout_destination == "queue":
-        stdout.put(None)
-    if stderr_destination == "queue":
-        stderr.put(None)
+    if not no_close_queues:
+        if stdout_destination == "queue":
+            stdout.put(None)
+        if stderr_destination == "queue":
+            stderr.put(None)
 
     # With polling, we return None if nothing has been send to the queues
     # With monitor, process.communicate() will result in '' even if nothing has been sent
@@ -959,10 +1261,28 @@ def command_runner(
     ):
         output_stderr = None
 
+    stdout_output = to_encoding(output_stdout, error_encoding, errors)
+    if stdout_output:
+        logger.debug("STDOUT: " + stdout_output)
+    if stderr_destination not in ["stdout", None]:
+        stderr_output = to_encoding(output_stderr, error_encoding, errors)
+        if stderr_output and not silent:
+            if exit_code == 0 or (
+                valid_exit_codes
+                and valid_exit_codes is True
+                or exit_code in valid_exit_codes
+            ):
+                logger.debug("STDERR: " + stderr_output)
+            else:
+                logger.error("STDERR: " + stderr_output)
+
+    if on_exit:
+        logger.debug("Running on_exit callable.")
+        on_exit()
+
     if split_streams:
         return exit_code, output_stdout, output_stderr
-    else:
-        return exit_code, _get_error_output(output_stdout, output_stderr)
+    return exit_code, _get_error_output(output_stdout, output_stderr)
 
 
 if sys.version_info[0] >= 3:
@@ -984,7 +1304,7 @@ def deferred_command(command, defer_time=300):
     seconds after it finished
     """
     # Use ping as a standard timer in shell since it's present on virtually *any* system
-    if os.name == "nt":
+    if os_name == "nt":
         deferrer = "ping 127.0.0.1 -n {} > NUL & ".format(defer_time)
     else:
         deferrer = "sleep {} && ".format(defer_time)
